/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.2 Baustellensimulation
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.simbaustelle.prognose;

import java.text.DateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;

import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;

import com.bitctrl.Constants;

import de.bsvrz.ibv.simbaustelle.simbaustelle.TestBaustelle;
import de.bsvrz.ibv.simbaustelle.simbaustelle.TestParameter;
import de.bsvrz.sys.funclib.bitctrl.util.UndefiniertException;

/**
 * Test für die korrekte Ermittlung des Berechnungszeitraums für die
 * Stauprognose einer Baustelle.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public class BerechnungsZeitraumTest {

	/** die Stunde zu der die Baustelle beendet wird. */
	private static final int BAUSTELLEN_ENDE_STUNDE = 14;

	/** die Stunde zu der die Baustelle begonnen wird. */
	private static final int BAUSTELLEN_START_STUNDE = 10;

	/** der Tag der Schließung der Baustelle. */
	private static final int BAUSTELLEN_ENDE_TAG = 5;

	/** der Tag der Eröffnung der Baustelle. */
	private static final int BAUSTELLEN_START_TAG = 4;

	/** das Jahr, in dem die Baustelle in Betrieb ist. */
	private static final int BAUSTELLE_JAHR = Calendar.getInstance().get(Calendar.YEAR) + 1;

	/** die Parameter für die Bestimmung des Berechnungszeitraums. */
	private final TestParameter parameter = new TestParameter();

	/**
	 * die Baustelle, für die der Berechnungszeitraum ermittelt werden soll.
	 */
	private final TestBaustelle baustelle = new TestBaustelle();

	/**
	 * interne Testfunktion zum überprüfen der Zeitbereichsdaten. Die Funktion
	 * erhält einen Startzeitpunkt für den der Zeitbereich erstellt werden soll,
	 * eine erwartete Anfangszeit und eine Anzahl der Intervalle, die für den
	 * ermittelten Zeitbereich erwartet werden
	 *
	 * @param startZeitpunkt
	 *            der Startzeitpunkt
	 * @param prognoseStart
	 *            der erwartete Prognosestart
	 * @param count
	 *            die Anzahl der prognoseintervalle
	 */
	private void intervallTest(final long startZeitpunkt, final long prognoseStart, final int count) {
		try {
			final BerechnungsZeitRaum zeitraum = new BerechnungsZeitRaum(startZeitpunkt, baustelle, parameter, false);
			Assert.assertEquals(count, zeitraum.getIntervalle().size());
			for (int idx = 0; idx < count; idx++) {
				final long erwartet = prognoseStart + ((idx * 2) * Constants.MILLIS_PER_HOUR);

				final long erhalten = zeitraum.getIntervalle().get(idx);
				Assert.assertEquals("Prognoseschritt " + idx,
						DateFormat.getDateTimeInstance().format(new Date(erwartet)),
						DateFormat.getDateTimeInstance().format(new Date(erhalten)));
			}

			Assert.assertTrue(zeitraum.isValid());

			if (zeitraum.getIntervalle().size() > 0) {
				Assert.assertEquals(prognoseStart, zeitraum.getStartZeitpunkt());
				Assert.assertEquals(zeitraum.getStartZeitpunkt() + (count * 2 * Constants.MILLIS_PER_HOUR),
						zeitraum.getEndeZeitpunkt());
			}

		} catch (final UndefiniertException e) {
			Assert.fail(e.getMessage());
		}
	}

	/**
	 * setzt den Zeitraum der Baustelle für den der Berechungszeitraum ermittelt
	 * werden soll. Die Baustelle wird am 4.5.2007 10:00 Uhr eröffnet und am
	 * 5.5.2007 14 Uhr geschlossen.
	 */
	@Before
	public final void setupBaustelle() {
		baustelle.setAnfang(new GregorianCalendar(BerechnungsZeitraumTest.BAUSTELLE_JAHR, Calendar.MAY,
				BerechnungsZeitraumTest.BAUSTELLEN_START_TAG, BerechnungsZeitraumTest.BAUSTELLEN_START_STUNDE, 0)
				.getTimeInMillis());
		baustelle.setEnde(new GregorianCalendar(BerechnungsZeitraumTest.BAUSTELLE_JAHR, Calendar.MAY,
				BerechnungsZeitraumTest.BAUSTELLEN_ENDE_TAG, BerechnungsZeitraumTest.BAUSTELLEN_ENDE_STUNDE, 0)
				.getTimeInMillis());
	}

	/**
	 * Testet die Erzeugung eines Berechungszeitraums mit den vorgegebenen
	 * Daten.
	 */
	@Test
	public final void testBerechnungsZeitRaum() {
		try {
			final BerechnungsZeitRaum berechnungsZeitRaum = new BerechnungsZeitRaum(System.currentTimeMillis(),
					baustelle, parameter, false);
			Assert.assertTrue(berechnungsZeitRaum != null);
		} catch (final UndefiniertException e) {
			Assert.fail(e.getMessage());
		}
	}

	/**
	 * Die Prognose findet am Tag der Baustelle nach dem Beginn der Baustelle
	 * statt. Es wird ein Zeitbereich von 0 Uhr, des Prognosetages mit 24
	 * Intervallen (mit je 2 Stunden) erwartet.
	 */
	@Test
	public final void testGetIntervalleBerechungszeitpunktAmTagDerBaustelleNachBeginn() {
		final long startZeitpunkt = new GregorianCalendar(BerechnungsZeitraumTest.BAUSTELLE_JAHR, Calendar.MAY,
				BerechnungsZeitraumTest.BAUSTELLEN_START_TAG, 12, 0).getTimeInMillis();
		final long prognoseStart = new GregorianCalendar(BerechnungsZeitraumTest.BAUSTELLE_JAHR, Calendar.MAY,
				BerechnungsZeitraumTest.BAUSTELLEN_START_TAG, 0, 0).getTimeInMillis();
		intervallTest(startZeitpunkt, prognoseStart, 24);
	}

	/**
	 * Die Prognose findet am Tag der Schließung der Baustelle statt.Es wird ein
	 * Zeitbereich von 0 Uhr, des Tages der Baustellenschließung mit 12
	 * Intervallen (mit je 2 Stunden) erwartet.
	 */
	@Test
	public final void testGetIntervalleBerechungszeitpunktAmTagDerBaustellenSchliessung() {
		final long startZeitpunkt = new GregorianCalendar(BerechnungsZeitraumTest.BAUSTELLE_JAHR, Calendar.MAY, 5, 12,
				0).getTimeInMillis();
		final long prognoseStart = new GregorianCalendar(BerechnungsZeitraumTest.BAUSTELLE_JAHR, Calendar.MAY, 5, 0, 0)
				.getTimeInMillis();
		intervallTest(startZeitpunkt, prognoseStart, 12);
	}

	/**
	 * Die Prognose findet am Tag der Baustelle vor dem Zeitpunkt des
	 * Baustellenbeginns statt. Es wird ein Zeitbereich von 0 Uhr, des
	 * Prognosetages mit 24 Intervallen (mit je 2 Stunden) erwartet.
	 */
	@Test
	public final void testGetIntervalleBerechungszeitpunktAmTagDerBaustelleVorBeginn() {
		final long startZeitpunkt = new GregorianCalendar(BerechnungsZeitraumTest.BAUSTELLE_JAHR, Calendar.MAY,
				BerechnungsZeitraumTest.BAUSTELLEN_START_TAG, 3, 0).getTimeInMillis();
		final long prognoseStart = new GregorianCalendar(BerechnungsZeitraumTest.BAUSTELLE_JAHR, Calendar.MAY,
				BerechnungsZeitraumTest.BAUSTELLEN_START_TAG, 0, 0).getTimeInMillis();
		intervallTest(startZeitpunkt, prognoseStart, 24);
	}

	/**
	 * Die Prognose findet nach dem Ende der Baustelle statt. Es wird ein
	 * Zeitbereich ohne Intervalle erwartet.
	 */
	@Test
	public final void testGetIntervalleBerechungszeitpunktNachDemBetriebDerBaustelle() {
		final long startZeitpunkt = new GregorianCalendar(BerechnungsZeitraumTest.BAUSTELLE_JAHR, Calendar.MAY, 6, 12,
				0).getTimeInMillis();
		final long prognoseStart = new GregorianCalendar(BerechnungsZeitraumTest.BAUSTELLE_JAHR, Calendar.MAY, 4, 0, 0)
				.getTimeInMillis();
		intervallTest(startZeitpunkt, prognoseStart, 0);
	}

	/**
	 * Die Prognose findet an einem Tag vor dem Start der Baustelle statt. Der
	 * erste Baustellentag liegt innerhalb des Prognosehorizonts. Es wird ein
	 * Zeitbereich von 0 Uhr, des Tages des Baustellenbeginns mit 12 Intervallen
	 * (mit je 2 Stunden) erwartet.
	 */
	@Test
	public final void testGetIntervalleBerechungszeitpunktVorBeginnDerBaustelleImPrognosehorizont() {
		final long startZeitpunkt = new GregorianCalendar(BerechnungsZeitraumTest.BAUSTELLE_JAHR, Calendar.MAY, 1, 12,
				0).getTimeInMillis();
		final long prognoseStart = new GregorianCalendar(BerechnungsZeitraumTest.BAUSTELLE_JAHR, Calendar.MAY, 4, 0, 0)
				.getTimeInMillis();
		intervallTest(startZeitpunkt, prognoseStart, 12);
	}

	/**
	 * Die Prognose findet an einem Tag vor dem Start der Baustelle statt. Die
	 * Baustelle liegt außerhalb des Prognosehorizonts. Es wird ein Zeitbereich
	 * ohne Intervalle erwartet.
	 */
	@Test
	public final void testGetIntervalleBerechungszeitpunktVorDemPrognosehorizont() {
		final long startZeitpunkt = new GregorianCalendar(BerechnungsZeitraumTest.BAUSTELLE_JAHR, Calendar.APRIL, 30,
				12, 0).getTimeInMillis();
		final long prognoseStart = new GregorianCalendar(BerechnungsZeitraumTest.BAUSTELLE_JAHR, Calendar.MAY, 4, 0, 0)
				.getTimeInMillis();
		intervallTest(startZeitpunkt, prognoseStart, 0);
	}

	/**
	 * Test für die Berechnung des Zeitraums einer manuellen simulierten
	 * Baustelle. Die Baustelle läuft vom 6.12.2022 10:30 bis zum 24.12.2022
	 * 12:22. Als Ergebnis wird der Bereich 6.12.2022 00:00 bis 25.12.2022 00:00
	 * erwartet.
	 *
	 * @throws UndefiniertException
	 *             ein Intervall konnte nicht erstellt werden, weil notwendige
	 *             Elemente nicht definiert sind
	 */
	@Test
	public final void testIntervallFuerManuelleSimulation() throws UndefiniertException {
		final long startZeitpunkt = new GregorianCalendar(2022, Calendar.DECEMBER, 6, 10, 30).getTimeInMillis();
		final long endeZeitpunkt = new GregorianCalendar(2022, Calendar.DECEMBER, 24, 12, 22).getTimeInMillis();

		final long prognoseStart = new GregorianCalendar(2022, Calendar.DECEMBER, 6, 0, 0).getTimeInMillis();
		final long prognoseEnde = new GregorianCalendar(2022, Calendar.DECEMBER, 25, 0, 0).getTimeInMillis();

		final TestBaustelle testBaustelle = new TestBaustelle();
		testBaustelle.setAnfang(startZeitpunkt);
		testBaustelle.setEnde(endeZeitpunkt);

		final BerechnungsZeitRaum berechnungsZeitRaum = new BerechnungsZeitRaum(System.currentTimeMillis(),
				testBaustelle, new TestParameter(), true);

		Assert.assertEquals("Anfangszeitraum bei manueller Simulation", prognoseStart,
				berechnungsZeitRaum.getStartZeitpunkt());
		Assert.assertEquals("Endezeitraum bei manueller Simulation", prognoseEnde,
				berechnungsZeitRaum.getEndeZeitpunkt());
	}

}
