/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.2 Baustellensimulation
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.simbaustelle.spez;

import java.util.Calendar;
import java.util.Collections;
import java.util.List;

import org.junit.After;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;

import com.bitctrl.Constants;

import de.bsvrz.dav.daf.main.ClientDavConnection;
import de.bsvrz.dav.daf.main.ClientDavParameters;
import de.bsvrz.dav.daf.main.CommunicationError;
import de.bsvrz.dav.daf.main.ConnectionException;
import de.bsvrz.dav.daf.main.InconsistentLoginException;
import de.bsvrz.dav.daf.main.MissingParameterException;
import de.bsvrz.dav.daf.main.config.ConfigurationChangeException;
import de.bsvrz.dav.daf.main.config.DynamicObjectType;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.bitctrl.modell.AnmeldeException;
import de.bsvrz.sys.funclib.bitctrl.modell.DatensatzUpdateEvent;
import de.bsvrz.sys.funclib.bitctrl.modell.DatensatzUpdateListener;
import de.bsvrz.sys.funclib.bitctrl.modell.DatensendeException;
import de.bsvrz.sys.funclib.bitctrl.modell.ObjektFactory;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.Baustelle;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.MessQuerschnitt;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.StrassenSegment;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.VerkehrModellNetz;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.onlinedaten.OdBaustellenSimulation;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.onlinedaten.OdBaustellenSimulation.Daten.StauEintrag;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.parameter.PdBaustellenEigenschaften;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.parameter.PdBaustellenSimulationModell;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.parameter.PdFundamentalDiagramm;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.parameter.PdSituationsEigenschaften;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.zustaende.BaustellenVeranlasser;

/**
 * Die Testfälle der Prüfspezifikation.
 *
 * @author BitCtrl Systems GmbH, Uwe Peuker
 */
public class SpezifikationIT implements DatensatzUpdateListener {

	/** die verwendete Datenverteilerverbindung. */
	private static ClientDavConnection connection;

	/** Mutex zur Synchronisation der Ergebnisauswertung. */
	private final Object mutex = new Object();

	/** das verwendete Netz. */
	private VerkehrModellNetz verkehrsNetz;

	/** die Baustelle, mit der die Simulation ausgeführt wird. */
	private Baustelle baustelle;

	/** markiert den Erfolg der Testausführung. */
	private boolean testOk;

	/**
	 * Ausführung des kompletten Testablaufs.
	 *
	 * @throws DatensendeException
	 *             es konnten nicht alle erforderlichen Testdaten an den
	 *             Datenverteiler versendet werden
	 * @throws AnmeldeException
	 *             es konnten nicht alle erforderlichen Anmeldungen beim
	 *             Datenverteiler ausgeführt werden
	 * @throws ConfigurationChangeException
	 *             die Testbaustelle konnte nicht angelegt werden
	 */
	@Test
	public synchronized void ablaufGesamt() throws DatensendeException, AnmeldeException, ConfigurationChangeException {
		setzeParameter();
		anlegenTestBaustelle();
		baustelle.simuliereBaustelle("TEST", "SWE6.2 Test");
		synchronized (mutex) {
			try {
				mutex.wait(2 * Constants.MILLIS_PER_MINUTE);
			} catch (final InterruptedException e) {
				Assert.fail(e.getMessage());
			}
		}

		Assert.assertEquals("Testergebnisse nicht korrekt", true, testOk);
	}

	/**
	 * legt die Testbaustelle an.
	 *
	 * @throws ConfigurationChangeException
	 *             die Baustelle konnte innerhalb der Konfiguration nicht
	 *             angelegt werden
	 * @throws DatensendeException
	 *             es konnten nicht alle erforderlichen Daten an den
	 *             Datenverteiler versendet werden
	 * @throws AnmeldeException
	 *             die Anmeldung für die erforderlichen Datenspezifikationen war
	 *             nicht möglich
	 */
	private void anlegenTestBaustelle() throws ConfigurationChangeException, DatensendeException, AnmeldeException {
		baustelle = (Baustelle) ObjektFactory.getInstanz().getModellobjekt("testbaustelle.swe6_2_test");
		final DynamicObjectType bstTyp = (DynamicObjectType) SpezifikationIT.connection.getDataModel()
				.getType("typ.baustelle");
		if (baustelle == null) {
			final SystemObject baustellenObjekt = SpezifikationIT.connection.getLocalConfigurationAuthority()
					.getConfigurationArea().createDynamicObject(bstTyp, "testbaustelle.swe6_2_test", "Testbaustelle");
			baustelle = (Baustelle) ObjektFactory.getInstanz().getModellobjekt(baustellenObjekt);
		}

		baustelle.getOnlineDatensatz(OdBaustellenSimulation.class)
		.addUpdateListener(OdBaustellenSimulation.Aspekte.PrognoseNormal.getAspekt(), this);

		final PdSituationsEigenschaften datenSatz = baustelle.getParameterDatensatz(PdSituationsEigenschaften.class);
		final Calendar now = Calendar.getInstance();
		now.set(Calendar.HOUR_OF_DAY, 10);
		now.set(Calendar.MINUTE, 0);
		now.set(Calendar.SECOND, 0);
		now.set(Calendar.MILLISECOND, 0);
		final PdSituationsEigenschaften.Daten daten = datenSatz.erzeugeDatum();
		daten.setStartZeit(now.getTimeInMillis() + Constants.MILLIS_PER_DAY);
		daten.setDauer((34 * Constants.MILLIS_PER_HOUR) - Constants.MILLIS_PER_MINUTE);
		daten.setSegmente(
				Collections.singletonList((StrassenSegment) ObjektFactory.getInstanz().getModellobjekt("ass.ass4")));
		daten.setStartOffset(0);
		daten.setEndOffset(0);

		datenSatz.anmeldenSender();
		datenSatz.sendeDaten(daten);
		datenSatz.abmeldenSender();

		// Baustellendaten setzen

		final PdBaustellenEigenschaften bstDatenSatz = baustelle.getParameterDatensatz(PdBaustellenEigenschaften.class);
		final PdBaustellenEigenschaften.Daten bstDaten = bstDatenSatz.erzeugeDatum();
		bstDaten.setRestKapazitaet(451);
		bstDaten.setVeranlasser(BaustellenVeranlasser.VRZ);
		bstDatenSatz.anmeldenSender();
		bstDatenSatz.sendeDaten(bstDaten);
		bstDatenSatz.abmeldenSender();

		verkehrsNetz.baustelleHinzufuegen(baustelle.getSystemObject());

	}

	@Override
	public void datensatzAktualisiert(final DatensatzUpdateEvent event) {

		final OdBaustellenSimulation datenSatz = (OdBaustellenSimulation) event.getDatensatz();
		if (datenSatz != null) {
			final OdBaustellenSimulation.Daten datum = datenSatz
					.getDatum(OdBaustellenSimulation.Aspekte.PrognoseNormal.getAspekt());
			if ((datum != null) && datum.isValid() && datum.isSimulationErfolgreich()) {

				final Calendar now = Calendar.getInstance();
				now.set(Calendar.HOUR_OF_DAY, 0);
				now.set(Calendar.MINUTE, 0);
				now.set(Calendar.SECOND, 0);
				now.set(Calendar.MILLISECOND, 0);
				final long testStart = now.getTimeInMillis() + Constants.MILLIS_PER_DAY;

				final List<StauEintrag> staus = datum.getStaus();
				Assert.assertEquals("Anzahl der Staus", 4, staus.size());

				StauEintrag stau = staus.get(0);
				Assert.assertEquals(testStart + (6L * Constants.MILLIS_PER_HOUR), stau.getStartZeit());
				Assert.assertEquals(28800L * Constants.MILLIS_PER_SECOND, stau.getDauer());
				Assert.assertEquals(3006L, stau.getMaxLaenge());
				Assert.assertEquals(testStart + (8L * Constants.MILLIS_PER_HOUR), stau.getMaxLaengeZeit());
				Assert.assertEquals(7229L, Math.round(stau.getVerlustZeit() / 1000.0));

				stau = staus.get(1);
				Assert.assertEquals(testStart + (16L * Constants.MILLIS_PER_HOUR), stau.getStartZeit());
				Assert.assertEquals(21600L * Constants.MILLIS_PER_SECOND, stau.getDauer());
				Assert.assertEquals(2346L, stau.getMaxLaenge());
				Assert.assertEquals(testStart + (18L * Constants.MILLIS_PER_HOUR), stau.getMaxLaengeZeit());
				Assert.assertEquals(6242L, Math.round(stau.getVerlustZeit() / 1000.0));

				stau = staus.get(2);
				Assert.assertEquals(testStart + (30L * Constants.MILLIS_PER_HOUR), stau.getStartZeit());
				Assert.assertEquals(28800L * Constants.MILLIS_PER_SECOND, stau.getDauer());
				Assert.assertEquals(3462L, stau.getMaxLaenge());
				Assert.assertEquals(testStart + (32L * Constants.MILLIS_PER_HOUR), stau.getMaxLaengeZeit());
				Assert.assertEquals(9212L, Math.round(stau.getVerlustZeit() / 1000.0));

				stau = staus.get(3);
				Assert.assertEquals(testStart + (40L * Constants.MILLIS_PER_HOUR), stau.getStartZeit());
				Assert.assertEquals(21600L * Constants.MILLIS_PER_SECOND, stau.getDauer());
				Assert.assertEquals(2262L, stau.getMaxLaenge());
				Assert.assertEquals(testStart + (42L * Constants.MILLIS_PER_HOUR), stau.getMaxLaengeZeit());
				Assert.assertEquals(6019L, Math.round(stau.getVerlustZeit() / 1000.0));

				testOk = true;
				synchronized (mutex) {
					mutex.notify();
				}

			}
		}

	}

	/** stellt die Verbindung zum Testdatenverteiler auf localhost:8083 her. */
	@Before
	public void datenverteilerVerbindung() {
		try {
			SpezifikationIT.connection = new ClientDavConnection(new ClientDavParameters());
			SpezifikationIT.connection.connect();
			SpezifikationIT.connection.login("Tester", "geheim");

			ObjektFactory.getInstanz().setVerbindung(SpezifikationIT.connection);
			ObjektFactory.getInstanz().registerStandardFactories();

			verkehrsNetz = (VerkehrModellNetz) ObjektFactory.getInstanz().getModellobjekt("netz.swe6Test");
			final GanglinienPrognoseProxy ganglinienPrognoseProxy = new GanglinienPrognoseProxy();
			Assert.assertTrue(ganglinienPrognoseProxy != null);

		} catch (final MissingParameterException e) {
			Assert.fail(e.getMessage());
		} catch (final CommunicationError e) {
			Assert.fail(e.getMessage());
		} catch (final ConnectionException e) {
			Assert.fail(e.getMessage());
		} catch (final InconsistentLoginException e) {
			Assert.fail(e.getMessage());
		}
	}

	/**
	 * setzt die für die jeweilige Testvariante vorgesehenen Parameter.
	 *
	 * @throws AnmeldeException
	 *             es konnten nicht alle erforderlichen Anmeldungen beim
	 *             Datenverteiler ausgeführt werden
	 * @throws DatensendeException
	 *             es konnten nicht alle erforderlichen Daten über den
	 *             Datenverteiler versendet werden
	 */
	private void setzeParameter() throws AnmeldeException, DatensendeException {
		final PdBaustellenSimulationModell datenSatz = verkehrsNetz
				.getParameterDatensatz(PdBaustellenSimulationModell.class);

		datenSatz.anmeldenSender();
		final PdBaustellenSimulationModell.Daten datum = datenSatz.erzeugeDatum();
		datum.setProgoseHorizont(3);
		datum.setAktualisierungsIntervall(3);
		datum.setIterationsSchrittweite(2 * Constants.MILLIS_PER_HOUR);
		datum.setFaktorQ0(50);
		datum.setLaengePkw(6);
		datum.setStauBeginn(360);
		datum.setStauAufhebung(350);

		datenSatz.sendeDaten(datum);
		datenSatz.abmeldenSender();

		// Fundamentaldiagramm für MQ4 setzen
		final MessQuerschnitt mq = (MessQuerschnitt) ObjektFactory.getInstanz().getModellobjekt("mq.MQ4");
		final PdFundamentalDiagramm fdDatenSatz = mq.getParameterDatensatz(PdFundamentalDiagramm.class);
		fdDatenSatz.anmeldenSender();
		final PdFundamentalDiagramm.Daten fdDatum = fdDatenSatz.erzeugeDatum();
		fdDatum.setQ0(998);
		fdDatum.setK0((short) 0);
		fdDatum.setV0((short) 0);
		fdDatum.setVFrei((short) 0);

		fdDatenSatz.sendeDaten(fdDatum);
		fdDatenSatz.abmeldenSender();
	}

	/**
	 * beendet die Datenverteilerverbindung nach Ausführung der Tests.
	 */
	@After
	public void verbindungBeenden() {
		SpezifikationIT.connection.disconnect(false, "Verbindung beendet");
	}
}
